import asyncio
from contextlib import AsyncExitStack
from typing import Any

from mcp import ClientSession, StdioServerParameters
from mcp import Tool as MCPTool
from mcp.client.stdio import stdio_client
from mcp.types import CallToolResult, TextContent
from textual import log

from oterm.config import appConfig
from oterm.types import Tool, ToolDefinition

mcp_clients = []


# adapted from mcp-python-sdk/examples/clients/simple-chatbot/mcp_simple_chatbot/main.py
class MCPClient:
    """Manages MCP server connections and tool execution."""

    def __init__(self, name: str, server_params: StdioServerParameters, errlog=None):
        self.name = name
        self.server_params = server_params
        self.errlog = errlog
        self.stdio_context: Any | None = None
        self.session: ClientSession | None = None
        self._cleanup_lock: asyncio.Lock = asyncio.Lock()
        self.exit_stack: AsyncExitStack = AsyncExitStack()

    async def initialize(self) -> None:
        """Initialize the server connection."""

        try:
            stdio_transport = await self.exit_stack.enter_async_context(
                stdio_client(self.server_params)
            )
            read, write = stdio_transport
            session = await self.exit_stack.enter_async_context(
                ClientSession(read, write)
            )
            await session.initialize()
            self.session = session
        except Exception as e:
            await self.cleanup()
            raise e

    async def get_available_tools(self) -> list[MCPTool]:
        """List available tools from the server.

        Returns:
            A list of available tools.

        Raises:
            RuntimeError: If the server is not initialized.
        """
        if not self.session:
            raise RuntimeError(f"Server {self.name} not initialized")

        tools_response = await self.session.list_tools()

        # Let's just ignore pagination for now
        return tools_response.tools

    async def call_tool(
        self,
        tool_name: str,
        arguments: dict[str, Any],
        retries: int = 2,
        delay: float = 1.0,
    ) -> Any:
        """Execute a tool with retry mechanism.

        Args:
            tool_name: Name of the tool to execute.
            arguments: Tool arguments.
            retries: Number of retry attempts.
            delay: Delay between retries in seconds.

        Returns:
            Tool execution result.

        Raises:
            RuntimeError: If server is not initialized.
            Exception: If tool execution fails after all retries.
        """
        if not self.session:
            raise RuntimeError(f"Server {self.name} not initialized")

        attempt = 0
        while attempt < retries:
            try:
                result = await self.session.call_tool(tool_name, arguments)
                return result

            except Exception as e:
                attempt += 1
                log.warning(
                    f"Error executing tool: {e}. Attempt {attempt} of {retries}."
                )
                if attempt < retries:
                    log.info(f"Retrying in {delay} seconds...")
                    await asyncio.sleep(delay)
                else:
                    log.error("Max retries reached. Failing.")
                    raise

    async def cleanup(self) -> None:
        """Clean up server resources."""
        async with self._cleanup_lock:
            try:
                await self.exit_stack.aclose()
                self.session = None
                self.stdio_context = None
            except Exception as e:
                log.error(f"Error during cleanup of MCP server {self.name}.")
                raise e


class MCPToolCallable:
    def __init__(self, name, server_name, client):
        self.name = name
        self.server_name = server_name
        self.client = client

    async def call(self, **kwargs):

        log.info(f"Calling Tool {self.name} in {self.server_name} with {kwargs}")
        res: CallToolResult = await self.client.call_tool(self.name, kwargs)
        if res.isError:
            log.error(f"Error call mcp tool {self.name}.")
            raise Exception(f"Error call mcp tool {self.name}.")
        text_content = [m.text for m in res.content if type(m) is TextContent]
        return "\n".join(text_content)


async def setup_mcp_servers():

    mcp_servers = appConfig.get("mcpServers")
    tool_defs: list[ToolDefinition] = []

    if mcp_servers:
        for server, config in mcp_servers.items():
            client = MCPClient(server, StdioServerParameters.model_validate(config))
            await client.initialize()
            mcp_clients.append(client)
            log.info(f"Initialized MCP server {server}")

            mcp_tools: list[MCPTool] = await client.get_available_tools()

            for mcp_tool in mcp_tools:
                tool = mcp_tool_to_ollama_tool(mcp_tool)
                mcpToolCallable = MCPToolCallable(mcp_tool.name, server, client)
                tool_defs.append({"tool": tool, "callable": mcpToolCallable.call})
                log.info(f"Loaded MCP tool {mcp_tool.name} from {server}")

    return tool_defs


async def teardown_mcp_servers():

    log.info("Tearing down MCP servers")
    # Important to tear down in reverse order
    mcp_clients.reverse()
    for client in mcp_clients:
        await client.cleanup()


def mcp_tool_to_ollama_tool(mcp_tool: MCPTool) -> Tool:
    """Convert an MCP tool to an Ollama tool"""

    return Tool(
        function=Tool.Function(
            name=mcp_tool.name,
            description=mcp_tool.description,
            parameters=Tool.Function.Parameters.model_validate(mcp_tool.inputSchema),
        ),
    )



from importlib import import_module
from typing import Awaitable, Callable, Sequence

from ollama._types import Tool

from oterm.config import appConfig
from oterm.types import ExternalToolDefinition, ToolDefinition


def load_tools(tool_defs: Sequence[ExternalToolDefinition]) -> Sequence[ToolDefinition]:
    tools = []
    for tool_def in tool_defs:
        tool_path = tool_def["tool"]

        try:
            module, tool = tool_path.split(":")
            module = import_module(module)
            tool = getattr(module, tool)
            if not isinstance(tool, Tool):
                raise Exception(f"Expected Tool, got {type(tool)}")
        except ModuleNotFoundError as e:
            raise Exception(f"Error loading tool {tool_path}: {str(e)}")

        callable_path = tool_def["callable"]
        try:
            module, function = callable_path.split(":")
            module = import_module(module)
            callable = getattr(module, function)
            if not isinstance(callable, (Callable, Awaitable)):
                raise Exception(f"Expected Callable, got {type(callable)}")
        except ModuleNotFoundError as e:
            raise Exception(f"Error loading callable {callable_path}: {str(e)}")
        tools.append({"tool": tool, "callable": callable})

    return tools


available: list[ToolDefinition] = []

external_tools = appConfig.get("tools")
if external_tools:
    available.extend(load_tools(external_tools))



from pathlib import Path

import pytest
from mcp import StdioServerParameters

from oterm.ollamaclient import OllamaLLM
from oterm.tools import MCPClient, MCPToolCallable
from oterm.types import Tool

mcp_server_executable = Path(__file__).parent / "mcp_servers.py"

server_config = {
    "oracle": {
        "command": "mcp",
        "args": ["run", str(mcp_server_executable.absolute())],
    }
}


@pytest.mark.asyncio
async def test_mcp():

    client = MCPClient(
        "oracle",
        StdioServerParameters.model_validate(server_config["oracle"]),
    )
    await client.initialize()
    tools = await client.get_available_tools()

    tool = tools[0]
    oterm_tool = Tool(
        function=Tool.Function(
            name=tool.name,
            description=tool.description,
            parameters=Tool.Function.Parameters.model_validate(tool.inputSchema),
        ),
    )

    mcpToolCallable = MCPToolCallable(tool.name, "oracle", client)
    llm = OllamaLLM(
        tool_defs=[{"tool": oterm_tool, "callable": mcpToolCallable.call}],
    )

    res = await llm.completion("Ask the oracle what is the best client for Ollama.")
    assert "oterm" in res
    await client.cleanup()



import asyncio
from pathlib import Path

import typer
from typing_extensions import Annotated

from oterm.command.create import app
from oterm.store.store import Store

cli = typer.Typer()


@cli.command("create")
def create(
    name: Annotated[str, typer.Argument(help="The name of the command to create.")]
):
    app.run(name)


@cli.command("list")
def list_commands():
    async def get_commands():
        store = await Store.get_store()
        commands = await store.get_chats(type="command")
        return commands

    commands = asyncio.run(get_commands())
    if not commands:
        typer.echo("No commands found.")
        return

    typer.echo("Commands found:")
    for command in commands:
        id, name, *rest = command
        path = Path.home() / ".local" / "bin" / name
        exists = path.exists()
        typer.echo(f"{id}: {name} -> {exists and str(path) or 'Not found'}")


@cli.command("delete")
def delete_command(
    id: Annotated[int, typer.Argument(help="The id of the command to delete.")]
):
    async def delete_comm():
        store = await Store.get_store()
        command = await store.get_chat(id=id)
        if not command:
            typer.echo("Command not found.")
            return
        _, name, *rest = command
        path = Path.home() / ".local" / "bin" / name
        path.unlink(missing_ok=True)
        await store.delete_chat(id=id)
        typer.echo("Command deleted.")

    asyncio.run(delete_comm())


if __name__ == "__main__":
    cli()



from mcp.server.fastmcp import Context, FastMCP

mcp = FastMCP("Oracle")


@mcp.resource("config://app")
def get_config() -> str:
    return "Oracle MCP server"


@mcp.tool()
async def oracle(query: str, ctx: Context) -> str:
    await ctx.report_progress(1, 2)
    return "Oracle says: oterm"



import json
import stat
from pathlib import Path
from typing import Iterable

from jinja2 import Environment, FileSystemLoader
from textual.app import App, ComposeResult, SystemCommand
from textual.binding import Binding
from textual.screen import Screen
from textual.widgets import Footer, Header, TabbedContent

from oterm.app.chat_edit import ChatEdit
from oterm.app.pull_model import PullModel
from oterm.app.splash import splash
from oterm.app.widgets.chat import ChatContainer
from oterm.config import appConfig
from oterm.store.store import Store
from oterm.tools.mcp import setup_mcp_servers, teardown_mcp_servers


class CreateCommandApp(App):
    TITLE = "oterm - Create Command"
    SUB_TITLE = "Create custom LLM chats as commands."
    CSS_PATH = "../app/oterm.tcss"
    BINDINGS = [
        Binding("ctrl+q", "quit", "quit", id="quit"),
    ]

    def get_system_commands(self, screen: Screen) -> Iterable[SystemCommand]:
        yield from super().get_system_commands(screen)

        yield SystemCommand(
            "Pull model",
            "Pulls (or updates) a model from the Ollama server",
            self.action_pull_model,
        )

    async def create_command(self) -> None:
        async def on_done(model_info) -> None:
            if model_info is None:
                await self.action_quit()
            model = json.loads(model_info)
            store = await Store.get_store()

            db_id = await store.save_chat(
                id=None,
                name=self.command_name,
                model=model["name"],
                system=model["system"],
                format=model["format"],
                parameters=model["parameters"],
                keep_alive=model["keep_alive"],
                tools=model["tools"],
                type="command",
            )
            # Load the template from the package
            environment = Environment(loader=FileSystemLoader(Path(__file__).parent))
            template = environment.get_template("command_template.py.jinja")
            path = Path.home() / ".local/bin" / self.command_name
            path.parent.mkdir(parents=True, exist_ok=True)
            with open(path, "w") as f:
                f.write(template.render(db_id=db_id, name=self.command_name))
            path.chmod(path.stat().st_mode | stat.S_IEXEC)
            await self.action_quit()

        await self.push_screen(ChatEdit(), callback=on_done)

    async def action_quit(self) -> None:
        self.log("Quitting...")
        await teardown_mcp_servers()
        return self.exit()

    async def action_pull_model(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            screen = PullModel("")
        else:
            chat = tabs.active_pane.query_one(ChatContainer)
            screen = PullModel(chat.ollama.model)
        self.push_screen(screen)

    async def load_mcp(self):
        from oterm.tools import available

        mcp_tool_defs = await setup_mcp_servers()
        available += mcp_tool_defs

    async def on_mount(self) -> None:
        theme = appConfig.get("theme")
        if theme:
            if theme == "dark":
                self.theme = "textual-dark"
            elif theme == "light":
                self.theme = "textual-light"
            else:
                self.theme = theme
        self.dark = appConfig.get("theme") == "dark"
        self.watch(self.app, "theme", self.on_theme_change, init=False)

        await self.load_mcp()

        async def on_splash_done(message) -> None:
            await self.create_command()

        if appConfig.get("splash-screen"):
            self.push_screen(splash, callback=on_splash_done)
        else:
            await on_splash_done("")

    def on_theme_change(self, old_value: str, new_value: str) -> None:
        if appConfig.get("theme") != new_value:
            appConfig.set("theme", new_value)

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()

    def run(self, name: str):
        self.command_name = name
        return super().run()


app = CreateCommandApp()
